# Copyright (C) 2010 by Dalen Bernaca
# YOU ARE ALLOWED TO USE AND DISTRIBUTE THIS PROGRAM FREELY
#
# YOU ARE ALLOWED TO MODIFY THIS PROGRAM IN ONE CONDITION:
# THAT YOU SEND YOUR MODIFIED VERSION TO THE AUTHOR BY E-MAIL
# WITH SHORT EXPLANATION REGARDING THE MODIFICATION(S)
# IF YOUR MODIFICATION(S) INCREASE THE QUALITY OF THE PROGRAM
# THEY WILL BE TAKEN IN VIEW FOR THE NEW OFFICIAL VERSION
#
# Author: Dalen Bernaca
#         dbernaca@blue-merlin.hr
# Version: 1.4XB
#
# THE AUTHOR IS NOT RESPONSIBLE FOR ANY DAMAGE INFLICTED BY MISSUSE/ABUSE OF THIS SOFTWARE
# ANY OTHER PROGRAM WITH SAME/SIMILAR NAME AND PURPOSE WHICH IS NOT PROCURED FROM
# http://www.brailleweb.com/
# IS NOT THE OFFICIAL VERSION WHATEVER MAY BE WRITTEN IN THE SOURCE CODE AND
# SHOULD BE TREATED WITH ENORMOUS CARE
# THE AUTHOR IS NOT ACCOUNTABLE FOR SUCH SOURCE CODE
#
# decoder.py is the result of author's knowledge, skill and creativity and a lot of asembled code legally optained from the internet.
# Thanks to all who put them there!

"""
decoder.py is a cross-platform module for decoding compressed audio files.
It uses external decoders by turning their stdout into file-like object which is completely compatible with wave.py module.

Usage:

>>> import decoder
>>>
>>> mp3 = decoder.open("something.mp3")
>>> print "Channels:", mp3.getnchannels()
Channels: 2
>>> print "Framerate:", mp3.getframerate()
Framerate: 44100
>>> print "Duration:", mp3.getnframes()/mp3.getframerate()
Duration: 206
>>>
>>> import wave
>>>
>>> wf = wave.open("something.wav", "w")
>>> wf.setnchannels(mp3.getnchannels())
>>> wf.setsampwidth(mp3.getsampwidth())
>>> wf.setnframes(mp3.getnframes())
>>> wf.setframerate(mp3.getframerate())
>>> data = mp3.readframes(1024)
>>> while data!="":
>>>     wf.writeframes(data)
>>>     data = mp3.readframes(1024)
>>> mp3.close()
>>> wf.close()
>>>

You can use the audiodev module or PyAudio library or sys.stdout or some external command line player like aplay in the same way as shown above.

FUNCTIONS:

open(name, fh=0) --> fakewave / wave.Wave_read / aifc.Aifc_read object
The name argument is the path to file you wish to decode.
Supported filetypes:
*.mp1, *.mp2, *.mp3
*.mp4, *.m4a, *.m4b, *.aac
*.ogg
*.wma (tested on windows only)
*.wav, *.aif (uncompressed files only)

fh (force header) argument can be one or zero or boolean and is used if an external decoder writes only raw data to stdout.
For instance, "faad" under Linux.

The open() returns an instance of a class with the following public methods:
      getnchannels()  -- returns number of audio channels (1 for
                         mono, 2 for stereo)
      getsampwidth()  -- returns sample width in bytes
      getframerate()  -- returns sampling frequency
      getnframes()    -- returns number of audio frames
      getcomptype()   -- returns compression type (always 'NONE' - linear samples)
      getcompname()   -- returns human-readable version of
                         compression type (always 'not compressed' linear samples)
      getparams()     -- returns a tuple consisting of all of the
                         above in the above order
      getmarkers()    -- returns None (for compatibility with the
                         aifc module)
                         The exception goes when aiff file is used.
      getmark(id)     -- raises an error since the mark does not
                         exist (for compatibility with the aifc module)
                         The exception goes when aiff file is used.
      readframes(n)   -- returns at most n frames of decoded audio
      rewind()        -- rewind to the beginning of the audio stream
      setpos(pos)     -- seek to the specified position
      tell()          -- return the current position
      close()         -- close the instance (make it unusable)
The position returned by tell() and the position given to setpos()
are compatible and have nothing to do with the actual position on stdout.
The close() method is called automatically when the class instance
is destroyed.

NOTA BENE:

The setpos() and rewind() are usually not accessible because of on-fly decoding and use of one of them often raises
the IOError: [Erno 29] Illegal seek.

Unfortunately, you cannot give the open read file pointer to open(), not yet.

getnframes() returns correct number of frames, but decoders nearly always cut off some unimportant frames and therefore be careful about this.
In this case fakewave.readframes() returns empty frames where the "" (empty string) would be returned if asking the external decoder.

CLASSES:

Error --> Exception raised when something is wrong with external decoder's output

fakewave --> Instance returned by open() function if compressed audio file given

DEPENDENCIES:

decoder.py depends on mutagen library - to calculate number of frames.

In order to make it work you need external decoders:
lame, faad, oggdec, wmadec.
To install them on Ubuntu/Debian use:
apt-get install lame
apt-get install vorbis-tools
apt-get install faad
wmadec has to be compiled from source (if possible at all)

For Windows:
As I hate searching for dependencies for years and then finding the wrong ones,
you will get them along in the same ZIP as decoder.py.
But better be warned that there might be newer versions somewhere on the internet.
That especially goes for "wmadec" which is still buggy.

In the "codecs.pdc" file you should put commands for external decoders.
How the system reaches them.

Unix common configuration:
lame=lame
faad=faad
oggdec=oggdec
wmadec=./codecs/wmadec

Windows preferable configuration:
lame=./codecs/lame.exe
oggdec=./codecs/oggdec.exe
faad=./codecs/faad.exe
wmadec=./codecs/wmadec.exe

TEST:

decoder.py is tested on:
Ubuntu Linux, Windows 98 SE and Windows XP

TODO:

When you use decoder.py on Windows 98, the console window opens regulary regardless the CREATE_NO_WINDOW creation flag.
This is Windows 9x natural behaviour, but I am sure the way arround must exists.
I would be thankful for any help regarding this "problem".
"""

from subprocess import *
from fileinfo import info
import os, struct, wave, aifc

if os.name=="nt":
    from win32con import CREATE_NO_WINDOW as noWindow
else: noWindow = 0

__author__ = "Dalen Bernaca"
__version__ = "1.4XB"

file = open("codecs.pdc", "r")
c = file.readlines()
file.close()
d = {}
for x in c:
    x = x.strip().split("=")
    d[x[0].lower()] = x[1]
lame = d["lame"]
faad = d["faad"]
oggdec = d["oggdec"]
wmadec = d["wmadec"]

class Error(Exception): pass

WAVEHEADER = 'RIFF$\x00\x00\x00WAVEfmt \x10\x00\x00\x00\x01\x00\x02\x00D\xac\x00\x00\x10\xb1\x02\x00\x04\x00\x10\x00data\x00\x00\x00\x00'
# The 2 channeled, 44100 hz, sample width 2, stereo WAVE header (0 frames)

class fakewave:
    _nframes = 0
    _nchannels = 0
    _sampwidth = 0
    _framerate = 0
    _comptype = "NONE"
    _compname = "not compressed"
    _soundpos = 0
    _framesize = 0

    """
    The following variables are added later and they don't exist in wav.Wave_read nor in aifc.Aifc_read.
    _chunksize
    _format
    _subchunk1size
    _audioformat
    _byterate
    _blockalign
    _bitspersample
    _dataloc
    These variables does not have their member access functions.
    """
    def __init__ (self, obj, fh=0):
        self.obj = obj
        self.forceheader = fh
        self.close = obj.stdout.close
        self.initfp()

    def __del__ (self):
        self.close()

    def initfp (self):
        """Extracts data in the first 44 bytes in a WAV stdout"""
        file = self.obj.stdout
        # Read in all data
        header = file.read(44)
        # Verify that the correct identifiers are present
        if (header[0:4] != "RIFF") or (header[12:16] != "fmt "):
            if self.forceheader: header = WAVEHEADER
            else: raise Error, "file does not start with RIFF id or fmt chunk missing"
        self._chunksize = struct.unpack('<L', header[4:8])[0]
        self._format = header[8:12]
        self._subchunk1size = struct.unpack('<L', header[16:20])[0]
        self._audioformat = struct.unpack('<H', header[20:22])[0]
        self._nchannels = struct.unpack('<H', header[22:24])[0]
        self._framerate = struct.unpack('<L', header[24:28])[0]
        self._byterate = struct.unpack('<L', header[28:32])[0]
        self._blockalign = struct.unpack('<H', header[32:34])[0]
        self._bitspersample = struct.unpack('<H', header[34:36])[0]
        self._sampwidth = (self._bitspersample+7)//8
        self._framesize = self._nchannels*self._sampwidth
        self._dataloc = header.find("data")
        if self._dataloc<=0: self._dataloc = self.locatedata()
        self._nframes = int(self._framerate*info(self.obj.filename).info.length)

    def locatedata (self):
        """Locate data chunk"""
        file = self.obj.stdout
        filesize = os.path.getsize(self.obj.filename)
        location = 0
        nextchunk = 12 # skip the RIFF header
        while 1:
            # Read subchunk header
            file.seek(nextchunk)
            header = file.read(8)
            if header[0:4] == "data":
                location = nextchunk
            nextchunk += (8 + struct.unpack('<L', header[4:8])[0])
            if nextchunk >= filesize: break
        return location

    def readframes (self, nframes):
        r = self.obj.stdout.read(nframes*self._framesize)
        if r=="" and self._soundpos+nframes<=self._nframes:
            r = (nframes*self._framesize)*"\x00"
        if r!="": self._soundpos += nframes
        return r

    def getnframes (self):
        return self._nframes

    def getnchannels (self):
        return self._nchannels

    def getframerate (self):
        return self._framerate

    def getsampwidth (self):
        return self._sampwidth

    def setpos (self, pos):
        pos = int(pos)
        l = self._nframes
        if pos < 0 or pos > l:
            raise Error, "position not in range"
        self.obj.stdout.seek(pos*self._framesize)
        self._soundpos = pos

    def tell (self):
        return self._soundpos

    def rewind (self):
        self.obj.stdout.seek(0)
        self._soundpos = 0

    def getmarkers (self): pass

    def getmark (self, id):
        raise Error, "no marks"

    def getcomptype (self):
        return self._comptype

    def getcompname (self):
        return self._compname

    def getparams (self):
        return self.getnchannels(), self.getsampwidth(), \
            self.getframerate(), self.getnframes(), \
            self.getcomptype(), self.getcompname()

    def getfp (self):
        return self.obj

def open (name, fh=0):
    name = os.path.normpath(name)
    if not os.path.exists(name):
        raise IOError, (2, "No such file or directory: '%s'" % name)
    ext = os.path.splitext(name)[1][1:].lower()
    if ext in ("mp4", "m4a", "m4b", "aac"):
        if os.name=="nt": cline = [faad, "-q", "-w", name]
        else: cline = [faad, "-q", "-f", "2", "-w", name]
        po = Popen(cline, shell=0, stdout=PIPE, stdin=PIPE, stderr=PIPE, creationflags=noWindow)
        po.stdin.close(); po.stderr.close()
        po.filename = name
        wf = fakewave(po, fh)
    elif ext=="ogg":
        if os.name=="nt": cline = [oggdec, "--stdout", name]
        else: cline = [oggdec, "-q", "-o", "-", name]
        po = Popen(cline, shell=0, stdout=PIPE, stdin=PIPE, stderr=PIPE, creationflags=noWindow)
        po.stdin.close(); po.stderr.close()
        po.filename = name
        wf = fakewave(po, fh)
    elif ext in ("wav", "wave"):
        wf = wave.open(name, "r")
    elif ext in ("aiff", "aif"):
        wf = aifc.open(name, "r")
    elif ext in ("mp1", "mp2", "mp3"):
        cline = [lame, "--quiet", "--decode", name, "-"]
        po = Popen(cline, shell=0, stdout=PIPE, stdin=PIPE, stderr=PIPE, creationflags=noWindow)
        po.stdin.close(); po.stderr.close()
        po.filename = name
        wf = fakewave(po, fh)
    elif ext=="wma":
        cline = [wmadec, "-w", name]
        po = Popen(cline, shell=0, stdout=PIPE, stdin=PIPE, stderr=PIPE, creationflags=noWindow)
        po.stdin.close()#; po.stderr.close()
        po.filename = name
        wf = fakewave(po, fh)
    else: wf = None
    return wf
